#include "player_character.hpp"
#include "utils/godot_macros.h"
#include <godot_cpp/classes/navigation_agent3d.hpp>
#include <godot_cpp/variant/utility_functions.hpp>

namespace godot {
void PlayerCharacter::_bind_methods() {
#define CLASSNAME PlayerCharacter
    GDPROPERTY_HINTED(rotation_speed_curve, Variant::OBJECT, PROPERTY_HINT_RESOURCE_TYPE, "Curve");
}

void PlayerCharacter::_enter_tree() { GDGAMEONLY();
    this->nav_agent = this->get_node<NavigationAgent3D>("NavigationAgent3D");
    this->target_rotation = this->get_global_transform().get_basis().get_quaternion();
}

void PlayerCharacter::_process(double delta_time) { GDGAMEONLY();
    this->process_rotation(delta_time);
    if(!this->mode_manual) {
        this->process_ai(delta_time);
    }
}

void PlayerCharacter::_physics_process(double delta_time) { GDGAMEONLY();
    Vector3 const new_velocity = this->get_velocity().move_toward(this->velocity_target, delta_time * PlayerCharacter::ACCELERATION);
    this->set_velocity(new_velocity);
    this->move_and_slide();
}

void PlayerCharacter::move(Vector3 world_vector) {
    this->velocity_target = world_vector * PlayerCharacter::WALK_SPEED;
}

void PlayerCharacter::aim(Vector3 at) {
    Vector3 const position{this->get_global_position()};
    Vector3 const forward{(Vector3{at.x, 0.f, at.z} - Vector3{position.x, 0.f, position.z}).normalized()};
    Vector3 const up{0.f, 1.f, 0.f};
    this->target_rotation = Basis{up.cross(forward), up, forward};
}

void PlayerCharacter::set_manual_mode(bool value) {
    this->mode_manual = value;
    this->nav_agent->set_process_mode(value ? ProcessMode::PROCESS_MODE_PAUSABLE : ProcessMode::PROCESS_MODE_DISABLED);
}

void PlayerCharacter::set_rotation_speed_curve(Ref<Curve> curve) {
    this->rotation_speed_curve = curve;
}

Ref<Curve> PlayerCharacter::get_rotation_speed_curve() const {
    return this->rotation_speed_curve;
}

void PlayerCharacter::process_ai(double delta_time) {
    this->velocity_target = this->nav_agent->get_velocity();
}

void PlayerCharacter::process_rotation(double delta_time) {
    Transform3D trans{this->get_global_transform()};
    Basis basis = trans.get_basis();
    Quaternion const current_quaternion = basis.get_rotation_quaternion();
    Quaternion const target_quaternion = this->target_rotation.get_rotation_quaternion();
    float const angle = current_quaternion.angle_to(target_quaternion);
    basis.set_quaternion(angle >= delta_time * PlayerCharacter::ROTATION_SPEED
            ? current_quaternion.slerp(target_quaternion, delta_time * (this->rotation_speed_curve->sample(angle) * PlayerCharacter::ROTATION_SPEED) / angle)
            : target_quaternion);
    trans.set_basis(basis);
    this->set_global_transform(trans);
}

float const PlayerCharacter::ACCELERATION{100.f};
float const PlayerCharacter::WALK_SPEED{3.f};
float const PlayerCharacter::SPRINT_SPEED{5.f};
float const PlayerCharacter::ROTATION_SPEED{10.f};
}
